import colander
from pyramid.events import subscriber

from kinto.core import resource, utils
from kinto.core.events import ResourceChanged, ACTIONS
from kinto.schema_validation import validate_from_bucket_schema_or_400, JSONSchemaMapping


class CollectionSchema(resource.ResourceSchema):
    schema = JSONSchemaMapping(missing=colander.drop)
    cache_expires = colander.SchemaNode(colander.Int(), missing=colander.drop)


@resource.register(name='collection',
                   collection_path='/buckets/{{bucket_id}}/collections',
                   record_path='/buckets/{{bucket_id}}/collections/{{id}}')
class Collection(resource.ShareableResource):
    schema = CollectionSchema
    permissions = ('read', 'write', 'record:create')

    def get_parent_id(self, request):
        bucket_id = request.matchdict['bucket_id']
        parent_id = utils.instance_uri(request, 'bucket', id=bucket_id)
        return parent_id

    def process_record(self, new, old=None):
        """Additional collection schema validation from bucket, if any."""
        new = super().process_record(new, old)

        # Remove internal and auto-assigned fields.
        internal_fields = (self.model.id_field,
                           self.model.modified_field,
                           self.model.permissions_field)
        validate_from_bucket_schema_or_400(new, resource_name="collection", request=self.request,
                                           ignore_fields=internal_fields)
        return new


@subscriber(ResourceChanged,
            for_resources=('collection',),
            for_actions=(ACTIONS.DELETE,))
def on_collections_deleted(event):
    """Some collections were deleted, delete records.
    """
    storage = event.request.registry.storage
    permission = event.request.registry.permission

    for change in event.impacted_records:
        collection = change['old']
        bucket_id = event.payload['bucket_id']
        parent_id = utils.instance_uri(event.request, 'collection',
                                       bucket_id=bucket_id,
                                       id=collection['id'])
        storage.delete_all(collection_id=None,
                           parent_id=parent_id,
                           with_deleted=False)
        storage.purge_deleted(collection_id=None,
                              parent_id=parent_id)
        permission.delete_object_permissions(parent_id + '/*')



import colander
from jsonschema import Draft4Validator, ValidationError, SchemaError, RefResolutionError, validate
from pyramid.settings import asbool

from kinto.core import utils
from kinto.core.errors import raise_invalid
from kinto.views import object_exists_or_404


class JSONSchemaMapping(colander.SchemaNode):
    def schema_type(self, **kw):
        return colander.Mapping(unknown='preserve')

    def deserialize(self, cstruct=colander.null):
        # Start by deserializing a simple mapping.
        validated = super().deserialize(cstruct)

        # In case it is optional in parent schema.
        if not validated or validated in (colander.null, colander.drop):
            return validated
        try:
            check_schema(validated)
        except ValidationError as e:
            self.raise_invalid(e.message)
        return validated


def check_schema(data):
    try:
        Draft4Validator.check_schema(data)
    except SchemaError as e:
        message = e.path.pop() + e.message
        raise ValidationError(message)


def validate_schema(data, schema, ignore_fields=[]):
    required_fields = [f for f in schema.get('required', []) if f not in ignore_fields]
    # jsonschema doesn't accept 'required': [] yet.
    # See https://github.com/Julian/jsonschema/issues/337.
    # In the meantime, strip out 'required' if no other fields are required.
    if required_fields:
        schema = {**schema, 'required': required_fields}
    else:
        schema = {f: v for f, v in schema.items() if f != 'required'}

    data = {f: v for f, v in data.items() if f not in ignore_fields}

    try:
        validate(data, schema)
    except ValidationError as e:
        if e.path:
            field = e.path[-1]
        elif e.validator_value:
            field = e.validator_value[-1]
        else:
            field = e.schema_path[-1]
        e.field = field
        raise e
    # Raise an error here if a reference in the schema doesn't resolve.
    # jsonschema doesn't provide schema validation checking upon creation yet,
    # it must be validated against data.
    # See https://github.com/Julian/jsonschema/issues/399
    # For future support https://github.com/Julian/jsonschema/issues/346.
    except RefResolutionError as e:
        raise e


def validate_from_bucket_schema_or_400(data, resource_name, request, ignore_fields=[]):
    """Lookup in the parent objects if a schema was defined for this resource.

    If the schema validation feature is enabled, if a schema is/are defined, and if the
    data does not validate it/them, then it raises a 400 exception.
    """
    settings = request.registry.settings
    schema_validation = 'experimental_collection_schema_validation'
    # If disabled from settings, do nothing.
    if not asbool(settings.get(schema_validation)):
        return

    bucket_id = request.matchdict["bucket_id"]
    bucket_uri = utils.instance_uri(request, 'bucket', id=bucket_id)
    buckets = request.bound_data.setdefault('buckets', {})
    if bucket_uri not in buckets:
        # Unknown yet, fetch from storage.
        bucket = object_exists_or_404(request,
                                      collection_id='bucket',
                                      parent_id='',
                                      object_id=bucket_id)
        buckets[bucket_uri] = bucket

    # Let's see if the bucket defines a schema for this resource.
    metadata_field = "{}:schema".format(resource_name)
    bucket = buckets[bucket_uri]
    if metadata_field not in bucket:
        return

    # Validate or fail with 400.
    schema = bucket[metadata_field]
    try:
        validate_schema(data, schema, ignore_fields=ignore_fields)
    except ValidationError as e:
        raise_invalid(request, name=e.field, description=e.message)
    except RefResolutionError as e:
        raise_invalid(request, name='schema', description=str(e))



import colander
from cornice import Service
from pyramid.security import NO_PERMISSION_REQUIRED


contribute = Service(name='contribute.json',
                     description='Open-source information',
                     path='/contribute.json')


class ContributeResponseSchema(colander.MappingSchema):
    body = colander.SchemaNode(colander.Mapping(unknown='preserve'))


contribute_responses = {
    '200': ContributeResponseSchema(
        description='Return open source contributing information.')
}


@contribute.get(permission=NO_PERMISSION_REQUIRED, tags=['Utilities'],
                operation_id='contribute', response_schemas=contribute_responses)
def contribute_get(request):
    return {
        'name': 'kinto',
        'description': 'A minimalist JSON storage service.',
        'repository': {
            'url': 'https://github.com/Kinto/kinto',
            'license': 'Apache License (2.0)'
        },
        'participate': {
            'docs': 'https://kinto.readthedocs.io/',
            'mailing-list': 'kinto@mozilla.org',
            'irc': 'irc://irc.freenode.net/#kinto',
        },
        'keywords': [
            'JSON',
            'Python',
            'Offline',
            'Sync',
            'Storage',
        ],
        'urls': {
            'dev': 'https://kinto.dev.mozaws.net/v1/'
        }
    }



import unittest

from kinto import main
from .support import BaseWebTest


class TestMain(BaseWebTest, unittest.TestCase):
    def test_init_sets_command_on_registry(self):
        app = main({'command': 'migrate'}, None, **self.get_app_settings())

        self.assertEqual(app.registry.command, 'migrate')

    def test_init_with_no_global_config_doesnt_crash(self):
        app = main(None, None, **self.get_app_settings())

        self.assertEqual(app.registry.command, None)



from kinto.core.testing import unittest

from .support import BaseWebTest


BUCKET_URL = '/buckets/blog'
GROUP_URL = '/buckets/blog/groups/tarnac-nine'


SCHEMA = {
    'title': 'Group schema',
    'type': 'object',
    'required': ['phone'],
    'properties': {
        'phone': {
            'type': 'string'
        },
        'members': {
            'type': 'array',
            'items': {
                'type': 'string'
            },
        },
    },
}

VALID_GROUP = {'members': ['oidc:coupat'], 'phone': '+33688776655'}


class DeactivatedSchemaTest(BaseWebTest, unittest.TestCase):
    def test_schema_should_be_json_schema(self):
        newschema = {**SCHEMA, 'type': 'Washmachine'}
        resp = self.app.put_json(BUCKET_URL, {'data': {'group:schema': newschema}},
                                 headers=self.headers, status=400)
        error_msg = "'Washmachine' is not valid under any of the given schemas"
        self.assertIn(error_msg, resp.json['message'])

    def test_group_are_not_invalid_if_do_not_match_schema(self):
        self.app.put_json(BUCKET_URL, {'data': {'group:schema': SCHEMA}},
                          headers=self.headers)
        self.app.put_json(GROUP_URL, {'data': {'phone': 42}},
                          headers=self.headers, status=201)


class BaseWebTestWithSchema(BaseWebTest):

    @classmethod
    def get_app_settings(cls, extras=None):
        settings = super().get_app_settings(extras)
        settings['experimental_collection_schema_validation'] = 'True'
        return settings


class CollectionValidationTest(BaseWebTestWithSchema, unittest.TestCase):
    def setUp(self):
        super().setUp()
        resp = self.app.put_json(BUCKET_URL,
                                 {'data': {'group:schema': SCHEMA}},
                                 headers=self.headers)
        self.collection = resp.json['data']

    def test_empty_metadata_can_be_validated(self):
        self.app.post_json(BUCKET_URL + '/groups',
                           headers=self.headers,
                           status=400)
        self.app.post_json(BUCKET_URL + '/groups',
                           {'data': {}},
                           headers=self.headers,
                           status=400)

    def test_groups_are_valid_if_match_schema(self):
        self.app.put_json(GROUP_URL,
                          {'data': VALID_GROUP},
                          headers=self.headers,
                          status=201)

    def test_groups_are_invalid_if_do_not_match_schema(self):
        self.app.put_json(GROUP_URL,
                          {'data': {'phone': 42}},
                          headers=self.headers,
                          status=400)


SCHEMA_UNRESOLVABLE = {
    'properties': {
        'phone': {'$ref': '#/definitions/phone'}
    },
}


class CollectionUnresolvableTest(BaseWebTestWithSchema, unittest.TestCase):
    def setUp(self):
        super().setUp()
        resp = self.app.put_json(BUCKET_URL,
                                 {'data': {'group:schema': SCHEMA_UNRESOLVABLE}},
                                 headers=self.headers)
        self.collection = resp.json['data']

    def test_unresolvable_errors_handled(self):
        self.app.put_json(GROUP_URL,
                          {'data': {'phone': 42}},
                          headers=self.headers, status=400)



import mock
import unittest

from kinto.core import utils
from kinto.core.testing import get_user_headers
from pyramid.exceptions import ConfigurationError

from kinto.plugins.accounts import scripts, ACCOUNT_CACHE_KEY
from .. import support


class AccountsWebTest(support.BaseWebTest, unittest.TestCase):

    @classmethod
    def get_app_settings(cls, extras=None):
        if extras is None:
            extras = {}
        extras.setdefault('multiauth.policies', 'account')
        extras.setdefault('includes', 'kinto.plugins.accounts')
        extras.setdefault('account_create_principals', 'system.Everyone')
        # XXX: this should be a default setting.
        extras.setdefault('multiauth.policy.account.use', 'kinto.plugins.accounts.authentication.'
                                                          'AccountsAuthenticationPolicy')
        extras.setdefault('account_cache_ttl_seconds', '30')
        return super().get_app_settings(extras)


class BadAccountsConfigTest(support.BaseWebTest, unittest.TestCase):

    def test_raise_configuration_if_accounts_not_mentioned(self):
        with self.assertRaises(ConfigurationError) as cm:
            self.make_app({
                'includes': 'kinto.plugins.accounts',
                'multiauth.policies': 'basicauth',
            })
        assert "Account policy missing" in str(cm.exception)


class HelloViewTest(AccountsWebTest):

    def test_accounts_capability_if_enabled(self):
        resp = self.app.get('/')
        capabilities = resp.json['capabilities']
        self.assertIn('accounts', capabilities)


class AccountCreationTest(AccountsWebTest):

    def test_anyone_can_create_an_account(self):
        self.app.post_json('/accounts', {'data': {'id': 'alice', 'password': '12éé6'}},
                           status=201)

    def test_account_can_be_created_with_put(self):
        self.app.put_json('/accounts/alice', {'data': {'password': '123456'}}, status=201)

    def test_password_is_stored_encrypted(self):
        self.app.put_json('/accounts/alice', {'data': {'password': '123456'}}, status=201)
        stored = self.app.app.registry.storage.get(parent_id='alice',
                                                   collection_id='account',
                                                   object_id='alice')
        assert stored['password'] != '123456'

    def test_authentication_is_accepted_if_account_exists(self):
        self.app.post_json('/accounts', {'data': {'id': 'me', 'password': 'bouh'}},
                           status=201)
        resp = self.app.get('/', headers=get_user_headers('me', 'bouh'))
        assert resp.json['user']['id'] == 'account:me'

    def test_password_field_is_mandatory(self):
        self.app.post_json('/accounts', {'data': {'id': 'me'}}, status=400)

    def test_id_field_is_mandatory(self):
        self.app.post_json('/accounts', {'data': {'password': 'pass'}}, status=400)

    def test_id_can_be_email(self):
        self.app.put_json('/accounts/alice@example.com', {'data': {'password': '123456'}},
                          status=201)

    def test_account_can_have_metadata(self):
        resp = self.app.post_json('/accounts',
                                  {'data': {'id': 'me', 'password': 'bouh', 'age': 42}},
                                  status=201)
        assert resp.json['data']['age'] == 42

    def test_cannot_create_account_if_already_exists(self):
        self.app.post_json('/accounts', {'data': {'id': 'me', 'password': 'bouh'}},
                           status=201)
        resp = self.app.post_json('/accounts', {'data': {'id': 'me', 'password': 'bouh'}},
                                  status=403)
        assert 'already exists' in resp.json['message']

    def test_username_and_account_id_must_match(self):
        resp = self.app.put_json('/accounts/alice', {'data': {'id': 'bob', 'password': 'bouh'}},
                                 status=400)
        assert 'does not match' in resp.json['message']

    def test_returns_existing_account_if_authenticated(self):
        self.app.post_json('/accounts', {'data': {'id': 'me', 'password': 'bouh'}},
                           status=201)
        self.app.post_json('/accounts', {'data': {'id': 'me', 'password': 'bouh'}},
                           headers=get_user_headers('me', 'bouh'),
                           status=200)

    def test_cannot_create_other_account_if_authenticated(self):
        self.app.post_json('/accounts', {'data': {'id': 'me', 'password': 'bouh'}},
                           status=201)
        resp = self.app.post_json('/accounts', {'data': {'id': 'you', 'password': 'bouh'}},
                                  headers=get_user_headers('me', 'bouh'),
                                  status=400)
        assert 'do not match' in resp.json['message']

    def test_authentication_does_not_call_bcrypt_twice(self):
        self.app.post_json('/accounts', {'data': {'id': 'me', 'password': 'bouh'}},
                           status=201)
        with mock.patch('kinto.plugins.accounts.authentication.bcrypt') as mocked_bcrypt:
            resp = self.app.get('/', headers=get_user_headers('me', 'bouh'))
            assert resp.json['user']['id'] == 'account:me'

            resp = self.app.get('/', headers=get_user_headers('me', 'bouh'))
            assert resp.json['user']['id'] == 'account:me'

            mocked_bcrypt.checkpw.assert_called_once()

    def test_authentication_checks_bcrypt_again_if_password_changes(self):
        self.app.post_json('/accounts', {'data': {'id': 'me', 'password': 'bouh'}},
                           status=201)
        with mock.patch('kinto.plugins.accounts.authentication.bcrypt') as mocked_bcrypt:
            resp = self.app.get('/', headers=get_user_headers('me', 'bouh'))
            assert resp.json['user']['id'] == 'account:me'

            self.app.patch_json('/accounts/me', {'data': {'password': 'blah'}},
                                status=200, headers=get_user_headers('me', 'bouh'))

            resp = self.app.get('/', headers=get_user_headers('me', 'blah'))
            assert resp.json['user']['id'] == 'account:me'

            assert mocked_bcrypt.checkpw.call_count == 2

    def test_authentication_refresh_the_cache_each_time_we_authenticate(self):
        hmac_secret = self.app.app.registry.settings['userid_hmac_secret']
        cache_key = utils.hmac_digest(hmac_secret, ACCOUNT_CACHE_KEY.format('me'))

        self.app.post_json('/accounts', {'data': {'id': 'me', 'password': 'bouh'}},
                           status=201)
        resp = self.app.get('/', headers=get_user_headers('me', 'bouh'))
        assert resp.json['user']['id'] == 'account:me'

        self.app.app.registry.cache.expire(cache_key, 10)

        resp = self.app.get('/', headers=get_user_headers('me', 'bouh'))
        assert resp.json['user']['id'] == 'account:me'

        assert self.app.app.registry.cache.ttl(cache_key) >= 20

        resp = self.app.get('/', headers=get_user_headers('me', 'blah'))
        assert 'user' not in resp.json


class AccountUpdateTest(AccountsWebTest):

    def setUp(self):
        self.app.put_json('/accounts/alice', {'data': {'password': '123456'}}, status=201)

    def test_password_can_be_changed(self):
        self.app.put_json('/accounts/alice', {'data': {'password': 'bouh'}},
                          headers=get_user_headers('alice', '123456'),
                          status=200)

    def test_authentication_with_old_password_is_denied_after_change(self):
        self.app.put_json('/accounts/alice', {'data': {'password': 'bouh'}},
                          headers=get_user_headers('alice', '123456'),
                          status=200)
        self.app.get('/accounts/alice', headers=get_user_headers('alice', '123456'),
                     status=401)

    def test_authentication_with_new_password_is_accepted_after_change(self):
        self.app.put_json('/accounts/alice', {'data': {'password': 'bouh'}},
                          headers=get_user_headers('alice', '123456'),
                          status=200)
        self.app.get('/accounts/alice', headers=get_user_headers('alice', 'bouh'))

    def test_username_and_account_id_must_match(self):
        resp = self.app.patch_json('/accounts/alice', {'data': {'id': 'bob', 'password': 'bouh'}},
                                   headers=get_user_headers('alice', '123456'),
                                   status=400)
        assert 'does not match' in resp.json['message']

    def test_cannot_patch_unknown_account(self):
        self.app.patch_json('/accounts/bob', {'data': {'password': 'bouh'}},
                            headers=get_user_headers('alice', '123456'),
                            status=403)

    def test_cannot_patch_someone_else_account(self):
        self.app.put_json('/accounts/bob', {'data': {'password': 'bob'}}, status=201)
        self.app.patch_json('/accounts/bob', {'data': {'password': 'bouh'}},
                            headers=get_user_headers('alice', '123456'),
                            status=403)

    def test_metadata_can_be_changed(self):
        resp = self.app.patch_json('/accounts/alice', {'data': {'age': 'captain'}},
                                   headers=get_user_headers('alice', '123456'))
        assert resp.json['data']['age'] == 'captain'


class AccountDeleteTest(AccountsWebTest):

    def setUp(self):
        self.app.put_json('/accounts/alice', {'data': {'password': '123456'}}, status=201)

    def test_account_can_be_deleted(self):
        self.app.delete('/accounts/alice', headers=get_user_headers('alice', '123456'))

    def test_authentication_is_denied_after_delete(self):
        self.app.delete('/accounts/alice', headers=get_user_headers('alice', '123456'))
        self.app.get('/accounts/alice', headers=get_user_headers('alice', '123456'),
                     status=401)


class AccountViewsTest(AccountsWebTest):

    def setUp(self):
        self.app.put_json('/accounts/alice', {'data': {'password': '123456'}}, status=201)
        self.app.put_json('/accounts/bob', {'data': {'password': 'azerty'}}, status=201)

    def test_account_list_is_forbidden_if_anonymous(self):
        self.app.get('/accounts', status=401)

    def test_account_detail_is_forbidden_if_anonymous(self):
        self.app.get('/accounts/alice', status=401)

    def test_accounts_list_contains_only_one_record(self):
        resp = self.app.get('/accounts', headers=get_user_headers('alice', '123456'))
        assert len(resp.json['data']) == 1

    def test_account_record_can_be_obtained_if_authenticated(self):
        self.app.get('/accounts/alice', headers=get_user_headers('alice', '123456'))

    def test_cannot_obtain_someone_else_account(self):
        self.app.get('/accounts/bob', headers=get_user_headers('alice', '123456'),
                     status=403)

    def test_cannot_obtain_unknown_account(self):
        self.app.get('/accounts/jeanine', headers=get_user_headers('alice', '123456'),
                     status=403)


class PermissionsEndpointTest(AccountsWebTest):

    @classmethod
    def get_app_settings(cls, extras=None):
        settings = super().get_app_settings(extras)
        settings['experimental_permissions_endpoint'] = 'True'
        return settings

    def setUp(self):
        self.app.put_json('/accounts/alice', {'data': {'password': '123456'}}, status=201)
        self.headers = get_user_headers('alice', '123456')
        self.app.put('/buckets/a', headers=self.headers)
        self.app.put('/buckets/a/collections/b', headers=self.headers)
        self.app.put('/buckets/a/collections/b/records/c', headers=self.headers)

    def test_permissions_endpoint_is_compatible_with_accounts_plugin(self):
        resp = self.app.get('/permissions', headers=self.headers)
        uris = [r["uri"] for r in resp.json["data"]]
        assert uris == ['/buckets/a/collections/b/records/c',
                        '/buckets/a/collections/b',
                        '/buckets/a',
                        '/accounts/alice',
                        '/']

    def test_account_create_read_write_permissions_(self):
        resp = self.app.get('/permissions', headers=self.headers)
        buckets = resp.json['data']
        account = buckets[3]
        account['permissions'].sort()
        root = buckets[4]
        root['permissions'].sort()
        self.assertEqual(account, {
            'account_id': 'alice',
            'id': 'alice',
            'permissions': ['read', 'write'],
            'resource_name': 'account',
            'uri': '/accounts/alice'})
        self.assertEqual(root, {
            'permissions': ['account:create', 'bucket:create'],
            'resource_name': 'root',
            'uri': '/'})


class PermissionsEndpointTestUnauthenticatedCreatePermission(AccountsWebTest):

    @classmethod
    def get_app_settings(cls, extras=None):
        settings = super().get_app_settings(extras)
        settings['experimental_permissions_endpoint'] = 'True'
        settings['bucket_create_principals'] = 'system.Everyone'
        return settings

    def setUp(self):
        self.everyone_headers = get_user_headers('')

    def test_permissions_endpoint_is_compatible_with_accounts_plugin(self):
        resp = self.app.get('/permissions', headers=self.everyone_headers)
        buckets = resp.json['data']
        root = buckets[0]
        root['permissions'].sort()
        self.assertEqual(root, {
            'permissions': ['account:create', 'bucket:create'],
            'resource_name': 'root',
            'uri': '/'})


class AdminTest(AccountsWebTest):

    @classmethod
    def get_app_settings(cls, extras=None):
        settings = super().get_app_settings(extras)
        settings['account_write_principals'] = 'account:admin'
        settings['account_read_principals'] = 'account:admin'
        return settings

    def setUp(self):
        self.app.put_json('/accounts/admin', {'data': {'password': '123456'}}, status=201)
        self.admin_headers = get_user_headers('admin', '123456')

        self.app.put_json('/accounts/bob', {'data': {'password': '987654'}}, status=201)

    def test_admin_can_get_other_accounts(self):
        self.app.put_json('/accounts/alice', {'data': {'password': 'azerty'}}, status=201)

        resp = self.app.get('/accounts/alice', headers=get_user_headers('alice', 'azerty'))
        assert resp.json['data']['id'] == 'alice'

    def test_admin_can_delete_other_accounts(self):
        self.app.delete_json('/accounts/bob', headers=self.admin_headers)
        self.app.get('/accounts/bob', headers=get_user_headers('bob', '987654'),
                     status=401)

    def test_admin_can_set_others_password(self):
        self.app.patch_json('/accounts/bob', {'data': {'password': 'bouh'}},
                            headers=self.admin_headers)
        self.app.get('/accounts/bob', headers=get_user_headers('bob', '987654'),
                     status=401)
        self.app.get('/accounts/bob', headers=get_user_headers('bob', 'bouh'))

    def test_admin_can_retrieve_accounts_list(self):
        self.app.put_json('/accounts/alice', {'data': {'password': 'azerty'}}, status=201)

        resp = self.app.get('/accounts', headers=self.admin_headers)
        usernames = [r['id'] for r in resp.json['data']]
        assert sorted(usernames) == ['admin', 'alice', 'bob']

    def test_user_created_by_admin_with_put_can_see_her_record(self):
        self.app.put_json('/accounts/alice',
                          {'data': {'password': 'bouh'}},
                          headers=self.admin_headers)

        resp = self.app.get('/accounts/alice', headers=get_user_headers('alice', 'bouh'))
        assert resp.json['permissions'] == {'write': ['account:alice']}

    def test_user_created_by_admin_with_post_can_see_her_record(self):
        self.app.post_json('/accounts',
                           {'data': {'id': 'alice', 'password': 'bouh'}},
                           headers=self.admin_headers)

        resp = self.app.get('/accounts/alice', headers=get_user_headers('alice', 'bouh'))
        assert resp.json['permissions'] == {'write': ['account:alice']}

    def test_user_created_by_admin_can_delete_her_record(self):
        self.app.post_json('/accounts',
                           {'data': {'id': 'alice', 'password': 'bouh'}},
                           headers=self.admin_headers)

        self.app.delete('/accounts/alice', headers=get_user_headers('alice', 'bouh'))


class CheckAdminCreateTest(AccountsWebTest):
    def test_raise_if_create_but_no_write(self):
        with self.assertRaises(ConfigurationError):
            self.make_app({'account_create_principals': 'account:admin'})


class CreateOtherUserTest(AccountsWebTest):
    def setUp(self):
        self.app.put_json('/accounts/bob', {'data': {'password': '123456'}}, status=201)
        self.bob_headers = get_user_headers('bob', '123456')

    def test_create_other_id_is_still_required(self):
        self.app.post_json('/accounts', {'data': {'password': 'azerty'}}, status=400,
                           headers=self.bob_headers)

    def test_create_other_forbidden_without_write(self):
        self.app.put_json('/accounts/alice', {'data': {'password': 'azerty'}}, status=400,
                          headers=self.bob_headers)


class WithBasicAuthTest(AccountsWebTest):

    @classmethod
    def get_app_settings(cls, extras=None):
        if extras is None:
            extras = {'multiauth.policies': 'account basicauth'}
        return super().get_app_settings(extras)

    def test_password_field_is_mandatory(self):
        self.app.post_json('/accounts', {'data': {'id': 'me'}}, status=400)

    def test_id_field_is_mandatory(self):
        self.app.post_json('/accounts', {'data': {'password': 'pass'}}, status=400)

    def test_fallsback_on_basicauth(self):
        self.app.post_json('/accounts', {'data': {'id': 'me', 'password': 'bleh'}})

        resp = self.app.get('/', headers=get_user_headers('me', 'wrong'))
        assert 'basicauth' in resp.json['user']['id']

        resp = self.app.get('/', headers=get_user_headers('me', 'bleh'))
        assert 'account' in resp.json['user']['id']

    def test_raise_configuration_if_wrong_error(self):
        with self.assertRaises(ConfigurationError):
            self.make_app({'multiauth.policies': 'basicauth account'})


class CreateUserTest(unittest.TestCase):
    def setUp(self):
        self.registry = mock.MagicMock()
        self.registry.settings = {
            "includes": "kinto.plugins.accounts",
        }

    def test_create_user_in_read_only_displays_an_error(self):
        with mock.patch('kinto.plugins.accounts.scripts.logger') as mocked:
            self.registry.settings['readonly'] = 'true'
            code = scripts.create_user({'registry': self.registry})
            assert code == 51
            mocked.error.assert_called_once_with('Cannot create a user with '
                                                 'a readonly server.')

    def test_create_user_when_not_included_displays_an_error(self):
        with mock.patch('kinto.plugins.accounts.scripts.logger') as mocked:
            self.registry.settings['includes'] = ''
            code = scripts.create_user({'registry': self.registry})
            assert code == 52
            mocked.error.assert_called_once_with('Cannot create a user when the accounts '
                                                 'plugin is not installed.')

    def test_create_user_with_an_invalid_username_and_password_confirmation_recovers(self):
        with mock.patch('kinto.plugins.accounts.scripts.input', side_effect=["&zert", "username"]):
            with mock.patch('kinto.plugins.accounts.scripts.getpass.getpass', side_effect=[
                    "password", "p4ssw0rd", "password", "password"]):
                code = scripts.create_user({'registry': self.registry}, None, None)
                self.registry.storage.update.assert_called_once()
                self.registry.permission.add_principal_to_ace.assert_called_with(
                    "/accounts/username", "write", "account:username")
                assert code == 0

    def test_create_user_with_a_valid_username_and_password_confirmation(self):
        with mock.patch('kinto.plugins.accounts.scripts.input', return_value="username"):
            with mock.patch('kinto.plugins.accounts.scripts.getpass.getpass',
                            return_value="password"):
                code = scripts.create_user({'registry': self.registry})
                self.registry.storage.update.assert_called_once()
                self.registry.permission.add_principal_to_ace.assert_called_with(
                    "/accounts/username", "write", "account:username")
                assert code == 0

    def test_create_user_with_valid_username_and_password_parameters(self):
        code = scripts.create_user({'registry': self.registry}, "username", "password")
        self.registry.storage.update.assert_called_once()
        self.registry.permission.add_principal_to_ace.assert_called_with(
            "/accounts/username", "write", "account:username")
        assert code == 0

    def test_create_user_aborted_by_eof(self):
        with mock.patch('kinto.plugins.accounts.scripts.input', side_effect=EOFError):
            code = scripts.create_user({'registry': self.registry})
            assert code == 53



"""
kinto.core.scripts: utilities to build admin scripts for kinto-based services
"""
import logging

import transaction as current_transaction
from pyramid.settings import asbool

from kinto.core.storage import exceptions as storage_exceptions
from kinto.plugins.quotas import scripts as quotas


logger = logging.getLogger(__name__)


def migrate(env, dry_run=False):
    """
    User-friendly frontend to run database migrations.
    """
    registry = env['registry']
    settings = registry.settings
    readonly_backends = ('storage', 'permission')
    readonly_mode = asbool(settings.get('readonly', False))

    for backend in ('cache', 'storage', 'permission'):
        if hasattr(registry, backend):
            if readonly_mode and backend in readonly_backends:
                message = ('Cannot migrate the {} backend while '
                           'in readonly mode.'.format(backend))
                logger.error(message)
            else:
                getattr(registry, backend).initialize_schema(dry_run=dry_run)


def delete_collection(env, bucket_id, collection_id):
    registry = env['registry']
    settings = registry.settings
    readonly_mode = asbool(settings.get('readonly', False))

    if readonly_mode:
        message = ('Cannot delete the collection while in readonly mode.')
        logger.error(message)
        return 31

    bucket = '/buckets/{}'.format(bucket_id)
    collection = '/buckets/{}/collections/{}'.format(bucket_id, collection_id)

    try:
        registry.storage.get(collection_id='bucket',
                             parent_id='',
                             object_id=bucket_id)
    except storage_exceptions.RecordNotFoundError:
        logger.error("Bucket '{}' does not exist.".format(bucket))
        return 32

    try:
        registry.storage.get(collection_id='collection',
                             parent_id=bucket,
                             object_id=collection_id)
    except storage_exceptions.RecordNotFoundError:
        logger.error("Collection '{}' does not exist.".format(collection))
        return 33

    deleted = registry.storage.delete_all(collection_id='record',
                                          parent_id=collection,
                                          with_deleted=False)
    if len(deleted) == 0:
        logger.info("No records found for '{}'.".format(collection))
    else:
        logger.info('{} record(s) were deleted.'.format(len(deleted)))

    registry.storage.delete(collection_id='collection',
                            parent_id=bucket,
                            object_id=collection_id,
                            with_deleted=False)
    logger.info("'{}' collection object was deleted.".format(collection))

    record = ('/buckets/{bucket_id}'
              '/collections/{collection_id}'
              '/records/{record_id}')

    registry.permission.delete_object_permissions(
        collection,
        *[record.format(bucket_id=bucket_id,
                        collection_id=collection_id,
                        record_id=r['id']) for r in deleted])
    logger.info('Related permissions were deleted.')

    current_transaction.commit()

    return 0


def rebuild_quotas(env, dry_run=False):
    """Administrative command to rebuild quota usage information.

    This command recomputes the amount of space used by all
    collections and all buckets and updates the quota records in the
    storage backend to their correct values. This can be useful when
    cleaning up after a bug like e.g.
    https://github.com/Kinto/kinto/issues/1226.
    """
    registry = env['registry']
    settings = registry.settings
    readonly_mode = asbool(settings.get('readonly', False))

    # FIXME: readonly_mode is not meant to be a "maintenance mode" but
    # rather used with a database user that has read-only permissions.
    # If we ever introduce a maintenance mode, we should maybe enforce
    # it here.
    if readonly_mode:
        message = ('Cannot rebuild quotas while in readonly mode.')
        logger.error(message)
        return 41

    if 'kinto.plugins.quotas' not in settings['includes']:
        message = ('Cannot rebuild quotas when quotas plugin is not installed.')
        logger.error(message)
        return 42

    quotas.rebuild_quotas(registry.storage, dry_run=dry_run)
    current_transaction.commit()
    return 0



import os
import unittest

from kinto.plugins.admin import views as admin_views
from ..support import BaseWebTest


admin_module = os.path.dirname(admin_views.__file__)
build_folder = os.path.join(admin_module, "build")
built_index = os.path.join(build_folder, "index.html")


class AdminViewTest(BaseWebTest, unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        # Simulate admin bundle without npm
        os.makedirs(build_folder, exist_ok=True)
        with open(built_index, "w") as f:
            f.write("<html><script/></html>")

    @classmethod
    def tearDownClass(cls):
        super().tearDownClass()
        try:
            os.remove(built_index)
        except FileNotFoundError:
            pass

    @classmethod
    def get_app_settings(self, extras=None):
        settings = super().get_app_settings(extras)
        settings['includes'] = 'kinto.plugins.admin'
        return settings

    def setUp(self):
        admin_views.admin_home_view.saved = None

    def test_capability_is_exposed(self):
        self.maxDiff = None
        resp = self.app.get('/')
        capabilities = resp.json['capabilities']
        self.assertIn('admin', capabilities)
        self.assertIn('version', capabilities['admin'])
        del capabilities['admin']['version']
        expected = {
            "description": "Serves the admin console.",
            "url": ("https://github.com/Kinto/kinto-admin/"),
        }
        self.assertEqual(expected, capabilities['admin'])

    def test_permission_endpoint_is_enabled_with_admin(self):
        resp = self.app.get('/')
        capabilities = resp.json['capabilities']
        assert 'permissions_endpoint' in capabilities

    def test_admin_index_cat_be_reached(self):
        self.maxDiff = None
        resp = self.app.get('/admin/')
        assert "html" in resp.body.decode('utf-8')

    def test_admin_redirect_without_trailing_slash(self):
        resp = self.app.get('/admin', status=307)
        self.assertTrue(resp.headers['location'].endswith('/admin/'))



import logging
import getpass

import transaction as current_transaction
from pyramid.settings import asbool

from .utils import hash_password
from .views import AccountIdGenerator


logger = logging.getLogger(__name__)


def create_user(env, username=None, password=None):
    """Administrative command to create a new user."""
    registry = env['registry']
    settings = registry.settings
    readonly_mode = asbool(settings.get('readonly', False))
    if readonly_mode:
        message = 'Cannot create a user with a readonly server.'
        logger.error(message)
        return 51

    if 'kinto.plugins.accounts' not in settings['includes']:
        message = 'Cannot create a user when the accounts plugin is not installed.'
        logger.error(message)
        return 52

    try:
        validator = AccountIdGenerator()
        if username is None:
            username = input('Username: ')
        while not validator.match(username):
            print('{} is not a valid username.')
            print('Username should match {0!r}, please try again.'.format(validator.regexp))
            username = input('Username: ')

        if password is None:
            while True:  # The user didn't entered twice the same password
                password = getpass.getpass('Please enter a password for {}: '.format(username))
                confirm = getpass.getpass('Please confirm the password: '.format(username))

                if password != confirm:
                    print('Sorry, passwords do not match, please try again.')
                else:
                    break
    except EOFError:
        print('User creation aborted')
        return 53

    print("Creating user '{}'".format(username))
    record = {'id': username, 'password': hash_password(password)}
    registry.storage.update(collection_id='account',
                            parent_id=username,
                            object_id=username,
                            record=record)
    registry.permission.add_principal_to_ace('/accounts/{}'.format(username),
                                             'write',
                                             'account:{}'.format(username))

    current_transaction.commit()

    return 0



import re

from kinto.authorization import PERMISSIONS_INHERITANCE_TREE
from pyramid.exceptions import ConfigurationError

ACCOUNT_CACHE_KEY = 'accounts:{}:verified'
ACCOUNT_POLICY_NAME = 'account'

DOCS_URL = "https://kinto.readthedocs.io/en/stable/api/1.x/accounts.html"


def includeme(config):
    config.add_api_capability(
        'accounts',
        description='Manage user accounts.',
        url='https://kinto.readthedocs.io/en/latest/api/1.x/accounts.html')

    config.scan('kinto.plugins.accounts.views')

    PERMISSIONS_INHERITANCE_TREE['root'].update({
        'account:create': {}
    })
    PERMISSIONS_INHERITANCE_TREE['account'] = {
        'write': {'account': ['write']},
        'read': {'account': ['write', 'read']}
    }

    settings = config.get_settings()

    # Check that the account policy is mentioned in config if included.
    accountClass = 'AccountsAuthenticationPolicy'
    policy = None
    for k, v in settings.items():
        m = re.match('multiauth\.policy\.(.*)\.use', k)
        if m:
            if v.endswith(accountClass):
                policy = m.group(1)

    if not policy:
        error_msg = ("Account policy missing the 'multiauth.policy.*.use' "
                     "setting. See {} in docs {}.").format(accountClass, DOCS_URL)
        raise ConfigurationError(error_msg)

    # Add some safety to avoid weird behaviour with basicauth default policy.
    auth_policies = settings['multiauth.policies']
    if 'basicauth' in auth_policies and policy in auth_policies:
        if auth_policies.index('basicauth') < auth_policies.index(policy):
            error_msg = ("'basicauth' should not be mentioned before '%s' "
                         "in 'multiauth.policies' setting.") % policy
            raise ConfigurationError(error_msg)

    # We assume anyone in account_create_principals is to create
    # accounts for other people.
    # No one can create accounts for other people unless they are an
    # "admin", defined as someone matching account_write_principals.
    # Therefore any account that is in account_create_principals
    # should be in account_write_principals too.
    creators = set(settings.get('account_create_principals', '').split())
    admins = set(settings.get('account_write_principals', '').split())
    cant_create_anything = creators.difference(admins)
    # system.Everyone isn't an account.
    cant_create_anything.discard('system.Everyone')
    if cant_create_anything:
        message = ('Configuration has some principals in account_create_principals '
                   'but not in account_write_principals. These principals will only be '
                   'able to create their own accounts. This may not be what you want.\n'
                   'If you want these users to be able to create accounts for other users, '
                   'add them to account_write_principals.\n'
                   'Affected users: {}'.format(list(cant_create_anything)))

        raise ConfigurationError(message)



"""
kinto.plugins.quotas.scripts: scripts to maintain quotas and fix them when they're broken
"""
import logging

from kinto.core.storage import Sort
from kinto.core.storage.utils import paginated
from .listener import BUCKET_QUOTA_OBJECT_ID, COLLECTION_QUOTA_OBJECT_ID
from .utils import record_size

logger = logging.getLogger(__name__)

OLDEST_FIRST = Sort('last_modified', 1)


def rebuild_quotas(storage, dry_run=False):
    for bucket in paginated(storage, collection_id='bucket',
                            parent_id='', sorting=[OLDEST_FIRST]):
        bucket_id = bucket['id']
        bucket_path = '/buckets/{}'.format(bucket['id'])
        bucket_collection_count = 0
        bucket_record_count = 0
        bucket_storage_size = record_size(bucket)

        for collection in paginated(storage, collection_id='collection',
                                    parent_id=bucket_path, sorting=[OLDEST_FIRST]):
            collection_info = rebuild_quotas_collection(storage, bucket_id, collection, dry_run)
            (collection_record_count, collection_storage_size) = collection_info
            bucket_collection_count += 1
            bucket_record_count += collection_record_count
            bucket_storage_size += collection_storage_size

        bucket_record = {
            'record_count': bucket_record_count,
            'storage_size': bucket_storage_size,
            'collection_count': bucket_collection_count,
        }
        if not dry_run:
            storage.update(collection_id='quota', parent_id=bucket_path,
                           object_id=BUCKET_QUOTA_OBJECT_ID, record=bucket_record)

        logger.info('Bucket {}. Final size: {} collections, {} records, {} bytes.'.format(
            bucket_id, bucket_collection_count, bucket_record_count, bucket_storage_size))


def rebuild_quotas_collection(storage, bucket_id, collection, dry_run=False):
    """Helper method for rebuild_quotas that updates a single collection."""
    collection_id = collection['id']
    collection_record_count = 0
    collection_storage_size = record_size(collection)
    collection_path = '/buckets/{}/collections/{}'.format(bucket_id, collection_id)
    for record in paginated(storage, collection_id='record',
                            parent_id=collection_path, sorting=[OLDEST_FIRST]):
        collection_record_count += 1
        collection_storage_size += record_size(record)

    logger.info('Bucket {}, collection {}. Final size: {} records, {} bytes.'.format(
        bucket_id, collection_id, collection_record_count, collection_storage_size))
    new_quota_info = {
        'record_count': collection_record_count,
        'storage_size': collection_storage_size,
    }
    if not dry_run:
        storage.update(collection_id='quota', parent_id=collection_path,
                       object_id=COLLECTION_QUOTA_OBJECT_ID, record=new_quota_info)
    return (collection_record_count, collection_storage_size)



import os

from kinto.core.decorators import cache_forever

HERE = os.path.dirname(__file__)


# Configured home page
@cache_forever
def admin_home_view(request):
    try:
        with open(os.path.join(HERE, 'build/index.html')) as f:
            page_content = f.read()
    except FileNotFoundError:  # pragma: no cover
        with open(os.path.join(HERE, 'public/help.html')) as f:
            page_content = f.read()
    return page_content



from cornice import Service
from pyramid.security import NO_PERMISSION_REQUIRED

from kinto.events import ServerFlushed


flush = Service(name='flush',
                description='Clear database content',
                path='/__flush__')


@flush.post(permission=NO_PERMISSION_REQUIRED)
def flush_post(request):
    request.registry.storage.flush()
    request.registry.permission.flush()
    request.registry.cache.flush()
    event = ServerFlushed(request)
    request.registry.notify(event)

    request.response.status = 202
    return {}


def includeme(config):
    config.add_api_capability(
        'flush_endpoint',
        description='The __flush__ endpoint can be used to remove '
                    'all data from all backends.',
        url='https://kinto.readthedocs.io/en/latest/api/1.x/flush.html'
    )
    config.add_cornice_service(flush)



import bcrypt


def hash_password(password):
    # Store password safely in database as str
    # (bcrypt.hashpw returns base64 bytes).
    pwd_str = password.encode(encoding='utf-8')
    hashed = bcrypt.hashpw(pwd_str, bcrypt.gensalt())
    return hashed.decode(encoding='utf-8')



import json
import logging
import unittest

import mock
from pyramid import testing

from kinto.core import DEFAULT_SETTINGS, initialization, JsonLogFormatter

from .support import BaseWebTest


class RequestSummaryTest(BaseWebTest, unittest.TestCase):
    def setUp(self):
        super().setUp()
        config = testing.setUp()
        config.registry.settings = DEFAULT_SETTINGS
        initialization.setup_logging(config)

        patch = mock.patch('kinto.core.initialization.summary_logger')
        self.mocked = patch.start()
        self.addCleanup(patch.stop)

    def logger_context(self):
        args, kwargs = self.mocked.info.call_args_list[-1]
        return kwargs['extra']

    def test_standard_info_is_bound(self):
        headers = {
            'User-Agent': 'Smith',
            **self.headers
        }
        self.app.get('/', headers=headers)
        event_dict = self.logger_context()
        self.assertEqual(event_dict['path'], '/v0/')
        self.assertEqual(event_dict['method'], 'GET')
        self.assertEqual(event_dict['code'], 200)
        self.assertEqual(event_dict['agent'], 'Smith')
        self.assertIsNotNone(event_dict['uid'])
        self.assertIsNotNone(event_dict['time'])
        self.assertIsNotNone(event_dict['t'])
        self.assertEqual(event_dict['errno'], 0)
        self.assertNotIn('lang', event_dict)
        self.assertNotIn('headers', event_dict)
        self.assertNotIn('body', event_dict)

    def test_userid_is_none_when_anonymous(self):
        self.app.get('/')
        event_dict = self.logger_context()
        self.assertNotIn('uid', event_dict)

    def test_lang_is_not_none_when_provided(self):
        self.app.get('/', headers={'Accept-Language': 'fr-FR'})
        event_dict = self.logger_context()
        self.assertEqual(event_dict['lang'], 'fr-FR')

    def test_agent_is_not_none_when_provided(self):
        self.app.get('/', headers={'User-Agent': 'webtest/x.y.z'})
        event_dict = self.logger_context()
        self.assertEqual(event_dict['agent'], 'webtest/x.y.z')

    def test_errno_is_specified_on_error(self):
        self.app.get('/unknown', status=404)
        event_dict = self.logger_context()
        self.assertEqual(event_dict['errno'], 111)

    def test_basic_authn_type_is_bound(self):
        app = self.make_app({'multiauth.policies': 'basicauth'})
        app.get('/mushrooms', headers={'Authorization': 'Basic bWF0OjE='})
        event_dict = self.logger_context()
        self.assertEqual(event_dict['authn_type'], 'basicauth')

    def test_headers_and_body_when_level_is_debug(self):
        self.mocked.level = logging.DEBUG
        body = b'{"boom": 1}'
        self.app.post('/batch', body, headers=self.headers, status=400)
        event_dict = self.logger_context()
        self.assertEqual(event_dict['headers'], {
            'Authorization': 'Basic bWF0OnNlY3JldA==',
            'Content-Length': '11',
            'Content-Type': 'application/json',
            'Host': 'localhost:80'})
        self.assertEqual(event_dict['body'], body)

        self.maxDiff = None

        responseBody = event_dict['response']['body']
        self.assertEqual(json.loads(responseBody.decode('utf-8'))['error'], 'Invalid parameters')
        responseHeaders = event_dict['response']['headers']
        self.assertEqual(sorted(responseHeaders.keys()), [
            'Access-Control-Expose-Headers',
            'Content-Length',
            'Content-Type',
            'X-Content-Type-Options'])


class BatchSubrequestTest(BaseWebTest, unittest.TestCase):
    def setUp(self):
        super().setUp()

        patch = mock.patch('kinto.core.views.batch.subrequest_logger')
        self.subrequest_mocked = patch.start()
        self.addCleanup(patch.stop)

        patch = mock.patch('kinto.core.initialization.summary_logger')
        self.summary_mocked = patch.start()
        self.addCleanup(patch.stop)

        headers = {**self.headers, 'User-Agent': 'readinglist'}
        body = {
            'requests': [{
                'path': '/unknown',
                'headers': {'User-Agent': 'foo'}
            }, {
                'path': '/unknown2'
            }]
        }
        self.app.post_json('/batch', body, headers=headers)

    def test_batch_global_request_is_preserved(self):
        args, kwargs = self.summary_mocked.info.call_args_list[-1]
        extra = kwargs['extra']
        self.assertEqual(extra['code'], 200)
        self.assertEqual(extra['path'], '/v0/batch')
        self.assertEqual(extra['agent'], 'readinglist')

    def test_batch_size_is_bound(self):
        args, kwargs = self.summary_mocked.info.call_args_list[-1]
        extra = kwargs['extra']
        self.assertEqual(extra['batch_size'], 2)

    def test_subrequests_are_not_logged_as_request_summary(self):
        self.assertEqual(self.summary_mocked.info.call_count, 1)

    def test_subrequests_are_logged_as_subrequest_summary(self):
        self.assertEqual(self.subrequest_mocked.info.call_count, 2)
        args, kwargs = self.subrequest_mocked.info.call_args_list[-1]
        extra = kwargs['extra']
        self.assertEqual(extra['path'], '/v0/unknown2')
        args, kwargs = self.subrequest_mocked.info.call_args_list[-2]
        extra = kwargs['extra']
        self.assertEqual(extra['path'], '/v0/unknown')


class JsonFormatterTest(unittest.TestCase):
    def test_logger_name(self):
        JsonLogFormatter.init_from_settings({'project_name': 'kintowe'})
        f = JsonLogFormatter()
        record = logging.LogRecord('app.log', logging.DEBUG, '', 0, 'coucou', (), None)
        result = f.format(record)
        logged = json.loads(result)
        self.assertEqual(logged['Logger'], 'kintowe')
        self.assertEqual(logged['Type'], 'app.log')
        # See https://github.com/mozilla/mozilla-cloud-services-logger/issues/2
        self.assertEqual(logged['Fields']['msg'], 'coucou')
